# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Plot functions for the sluyspy package."""

import matplotlib as _mpl               # Get matplotlib
import matplotlib.pyplot as _plt        # Get matplotlib.pyplot
import numpy.core as _np                # Get NumPy
from .cli import error as _error


def start_plot(ptype='both', hsize=None,vsize=None, dark_bg=False, xkcd=False, title='Python plot',
               fz=None,lw=None, plot3d=False, man_zorder=True, cblind=True):
    """Start a matplotlib.pyplot plot with a choice of my favourite options.
    
    Parameters:
      ptype (str):        Plot type: 'screen', 'file', 'both' (a compromise; default), 'square' or 'org'.
      hsize (float):      Horizontal size of the figure (pixels if >50, hectopixels otherwise); can overrule setting by ptype.
      vsize (float):      Vertical size of the figure (pixels if >50, hectopixels otherwise); can overrule setting by ptype.
    
      dark_bg (bool):     Use a dark background (optional, defaults to False).
      xkcd (bool):        Use XKCD style (optional, defaults to False).
      title (str):        Window title (optional, defaults to 'Python plot').
    
      fz (int):           Font size (optional, defaults to None -> 14 or 16).
      lw (int):           Line width (optional, defaults to None -> 1 or 2).
    
      plot3d (bool):      Start a 3D plot (optional, defaults to False -> 2D).
      map_zorder (bool):  Use manual zorder (defaults to True - opposed to pyplot default: compute).
      cblind (bool):      Choose a colour-blindness-aware colour palette (optional, defaults to True).
    
    Returns:
      (tuple):  Tuple (fig,ax) containing the figure and axis objects.
    
    Defaults:
      - screen: size 1920x1080, font 14, lw: 1 - aimed at my screen;
      - file:   size 1250x700,  font 14, lw: 2 - aimed for file -> LaTeX;
      - both:   size 1580x850,  font 16, lw: 2 - a compromise between the above;
      - square: size 850x850,   font 16, lw: 2 - a square plot;
      - org:    size 930x520,   font 12, lw: 1 - aimed at output in emacs Orgmode.
    """
    
    if cblind is True: _plt.style.use('tableau-colorblind10')
    
    if xkcd:  _plt.xkcd()                                   # Plot everything that follows in XKCD style
    if dark_bg:  _plt.style.use('dark_background')          # Invert colours
    if xkcd and dark_bg:
        _plt.rcParams['path.effects'] = [_mpl.patheffects.withStroke(linewidth=0)]  # Needed for XKCD style on a dark background
    
    # If size>50, assume pixels/"dots" are specified; convert to "inches":
    if hsize is not None and hsize>50: hsize /= 100  # default: 100 dpi
    if vsize is not None and vsize>50: vsize /= 100  # default: 100 dpi
    
    if ptype == 'screen':
        if hsize is None: hsize = 19.2  # 1920 pixels
        if vsize is None: vsize = 10.8  # 1080 pixels
        _mpl.rcParams.update({'font.size': 14})       # Set font size for all text, aimed at full screen display
    elif ptype == 'file':
        if hsize is None: hsize = 12.5  # 1250 pixels
        if vsize is None: vsize =  7.0  # 700 pixels
        _mpl.rcParams.update({'font.size': 14})       # Set font size for all text aimed at an electonic file
        _mpl.rcParams.update({'lines.linewidth': 2})  # Set default line width to 2
    elif ptype == 'both':
        if hsize is None: hsize = 15.8  # 1580 pixels
        if vsize is None: vsize =  8.5  # 850 pixels
        _mpl.rcParams.update({'font.size': 16})       # Set font size for all text: compromise screen visibility and report readability
        _mpl.rcParams.update({'lines.linewidth': 2})  # Set default line width to 2
    elif ptype == 'square':
        if hsize is None: hsize = 8.5  # 850 pixels
        if vsize is None: vsize = 8.5  # 850 pixels
        _mpl.rcParams.update({'font.size': 16})       # Set font size for all text: compromise screen visibility and report readability
    elif ptype == 'org':
        if hsize is None: hsize = 9.0  # 900 pixels
        if vsize is None: vsize = 5.0  # 500 pixels
        _mpl.rcParams.update({'font.size': 12})       # Set font size for all text: small
    else:
        _error('Unknown plot type: '+ptype+', aborting.')
    
    # Find unique plot title ("number") - doublets cause mixed plots!
    titlel = title  # Local version
    plotnum = 2
    while _plt.fignum_exists(titlel):
        titlel = title + ' ' + str(plotnum)
    
    if fz is not None: _mpl.rcParams.update({'font.size': fz})        # Set font size for all text
    if lw is not None: _mpl.rcParams.update({'lines.linewidth': lw})  # Set custom default line width
    fig = _plt.figure(figsize=(hsize,vsize), num=titlel)  # Custom size
    
    if plot3d:
        ax = fig.add_subplot(111, projection='3d', computed_zorder=(not man_zorder))  # Create a 3D axes object for the current figure
    else:
        ax = fig.add_subplot(111)                                # Create a 2D axes object for the current figure
        
    if xkcd and dark_bg:  ax.spines[:].set_color('white')        # Needed for XKCD style on a dark background
    
    return fig, ax


def finish_plot(fig,ax, file_name=None, title=None, xlbl=None,ylbl=None, legend=False, legend_loc='best',
                grid=True, logx=False,logy=False, logx_gfmt=True, logy_gfmt=True, tight=1, save=True,
                close=True):
    """Show the current figure on screen or save it to disc and close it.
    
    Parameters:
      fig (pyplot.figure):  Current Pyplot figure object.
      ax (pyplot axes):     Current Pyplot axes object.
      file_name (str):      Name for the output file (optional, defaults to None -> screen).
    
      title (str):          Text to use as plot title (optional, defaults to None).
      xlbl (str):           Text to use as label for the horizontal axis (optional, defaults to None).
      ylbl (str):           Text to use as label for the vertical axis (optional, defaults to None).
    
      legend (bool):        Show a legend (optional, defaults to False).
      legend_loc (str):     Location for the legend (optional, defaults to 'best').
      grid (bool):          Show a grid (optional, defaults to True).
    
      logx (bool):          Make the horizontal axis logarithmic (optional, defaults to False -> linear).
      logy (bool):          Make the vertical axis logarithmic (optional, defaults to False -> linear).
    
      logx_gfmt (bool):     Use non-scientific format where possible on the horizontal axis.  Changes 10^0 -> 1, but also 10^40 -> 1e+40
      logy_gfmt (bool):     Use non-scientific format where possible on the vertical axis.  Changes 10^0 -> 1, but also 10^40 -> 1e+40
    
      tight (int):          Tightness of the margins: 0: not at all, 1: some (default), 2: quite, 3: very!
    
      save (bool):          Actually save the plot (if not, allow further changes, and don't close!).
      close (bool):         Actually close the plot (if not, allow further changes).
    """
    
    # Plot labels:
    if title is not None: _plt.title(title)                       # Plot title
    if xlbl is not None:  ax.set_xlabel(xlbl)                     # Label the horizontal axis
    if ylbl is not None:  ax.set_ylabel(ylbl)                     # Label the vertical axis
    
    # Plot features:
    if legend: ax.legend(loc=legend_loc)                          # Create a legend
    if logx:
        ax.set_xscale('log')                                 # Logarithmic horizontal axis
        if logx_gfmt: _plt.gca().xaxis.set_major_formatter(_mpl.ticker.FormatStrFormatter('%0g'))  # Use non-scientific format where possible
    if logy:
        ax.set_yscale('log')                                 # Logarithmic vertical axis
        if logy_gfmt: _plt.gca().yaxis.set_major_formatter(_mpl.ticker.FormatStrFormatter('%0g'))  # Use non-scientific format where possible
    if grid: ax.grid(grid)                                        # Plot a grid
    
    # Tightness of the margins:
    if tight>0: fig.tight_layout()                                # Use narrow margins
    
    if (file_name is None) or (file_name == 'screen'):
        if file_name == 'screen': print("sluyspy.plot.finish_plot(): file_name = 'screen' is obsolescent and will be removed.  Use None instead.")
        show_ctrlc()
    else:
        if save:
            if tight>2:
                fig.savefig(file_name, bbox_inches='tight', pad_inches=0)  # Use very, very narrow margins!
            elif tight>1:
                fig.savefig(file_name, bbox_inches='tight')                # Use very narrow margins
            else:
                fig.savefig(file_name)                                     # Save figure
        else:
            close = False  # Don't close the plot if you didn't save it!
    
    if close: _plt.close(fig=fig)
    
    return


def hist_norm(obj, x, bins=None, range=None, density=False, weights=None, cumulative=False, bottom=None,
              histtype='bar', align='mid', orientation='vertical', rwidth=None, log=False, color=None, label=None,
              stacked=False, *, data=None, **kwargs):
    """Draw a histogram normalised by total number rather than probability density.
    
    Parameters:
      obj (pyplot object):  Pyplot object or axes object to plot to.
      x (array or sequence of arrays):  Data.
      bins (int or sequence or str):  Number of bins.  Default: rcParams["hist.bins"] (default: 10).
      range (tuple or None):  Bin range, default: None.
      density (bool):  Plot probability density.  Default: False.
      weights (array-like or None):  Weights.  Default: None.
      cumulative (bool or -1):  Plot a cumulative histogram.  Default: False.
      bottom (array-like, scalar, or None):  Location of the bottom of the bins.  Default: None.
      histtype (str):  Histogram type: {'bar', 'barstacked', 'step', 'stepfilled'}, default: 'bar'.
      align (str):  Bin alignment: {'left', 'mid', 'right'}, default: 'mid'.
      orientation (str):  Bin orientation:  {'vertical', 'horizontal'}, default: 'vertical'.
      rwidth (float or None):  The relative width of the bars as a fraction of the bin width, default: None.
      log (bool):  Set the histogram axis to a log scale, default: False.
      color (color or array-like of colors or None):  Color or sequence of colors for bins, one per dataset, default: None.
      label (str or None):  Legend label, default: None.
      stacked (bool):  Stack multiple data on top of each other, rather than side by side, default: False.
      data (indexable object, optional):  If given, the following parameters also accept a string s, which is interpreted as data[s] (unless this raises an exception): x, weights.
      kwargs (dict):  Keyword arguments to be passed to pyplot.quiver().
    
    Returns:
      (tuple):  Tuple consisting of (n, bin_edges, patches):
      
      - (array or list of arrays):  The values of the histogram bins.
      - (array): The edges of the bins. Length nbins + 1.
      - (BarContainer or list of Polygons):  (List of) container(s) of individual artists used to create the histogram(s).
    """
        
    weights = _np.ones_like(x) / len(x)  # Set all weights to 1/n to pass them to pyplot.hist()
    
    n, bin_edges, patches = obj.hist(x, bins=bins, range=range, density=density, weights=weights,
                                     cumulative=cumulative, bottom=bottom, histtype=histtype, align=align,
                                     orientation=orientation, rwidth=rwidth, log=log, color=color,
                                     label=label, stacked=stacked, data=data, **kwargs)
    
    return n, bin_edges, patches


def show_ctrlc():
    """Call pyplot.show() and catch Ctrl-C to exit gently."""
    
    try:
        _plt.show()
    except KeyboardInterrupt:
        print(' - Received keyboard interrupt whilst showing plot, aborting.')  # " - " to create some space after "^C"
        exit(0)
        
    return


def pause_ctrlc(interval):
    """Call pyplot.pause() and catch Ctrl-C to exit gently.
    
    Parameters:
      interval (float):  pause time in seconds.
    """
    
    try:
        _plt.pause(interval)
    except KeyboardInterrupt:
        print(' - Received keyboard interrupt whilst pausing plot, aborting.')  # " - " to create some space after "^C"
        exit(0)
        
    return


def arrow_head_between_points(ax, xx,yy, dx=0,dy=0, scale=1, **kwargs):
    """Plot an arrow head between two 2D points and try to scale properly.
    
    Parameters:
      ax (axes):      Axes object to plot on.
      xx (float):     Array with two x values, for arrow-head position and direction.
      yy (float):     Array with two y values, for arrow-head position and direction.
      dx (float):     Shift arrow head in x direction.
      dy (float):     Shift arrow head in y direction.
      scale (float):  Scale arrow head.
      kwargs (dict):  Keyword arguments to be passed to pyplot.quiver().
    """
    
    # Local scaling parameter: undo arrow-head scaling with length:
    lscale = 0.1 * scale / _np.sqrt((xx[1]-xx[0])**2 + (yy[1]-yy[0])**2)
    
    if len(xx)!=2: _error('arrow_head_between_points(): The input variable xx must have two elements')
    if len(yy)!=2: _error('arrow_head_between_points(): The input variable yy must have two elements')
    
    ax.quiver(xx[1]+dx, yy[1]+dy, xx[1]-xx[0], yy[1]-yy[0],
              angles='xy', scale_units='width', pivot='mid',
              scale=3/lscale, width=5e-3*lscale,
              headwidth=10*lscale, headlength=15*lscale, headaxislength=10*lscale,
              **kwargs)
    return


def arrow_between_points(ax, xx,yy, scale=1, zorder=2, **kwargs):
    """Plot an arrow between two 2D points and try to scale the shaft and head properly.
    
    Parameters:
      ax (axes):      Axes object to plot on.
      xx (float):     Array with two x values, for tail and head position.
      yy (float):     Array with two y values, for tail and head position.
      scale (float):  Scale arrow shaft and head.
      zorder (int):   Zorder for the arrow.  Zorder<2 places it below the grid.
      kwargs (dict):  Keyword arguments to be passed to pyplot.quiver().
    """
    
    # Local scaling parameter: undo arrow-head scaling with length:
    sscale = 0.01 * scale  # Shaft AND head
    hscale = 2.5           # Head only, on tip of sscale
    
    if len(xx)!=2: _error('arrow_between_points(): The input variable xx must have two elements')
    if len(yy)!=2: _error('arrow_between_points(): The input variable yy must have two elements')
    
    ax.quiver(xx[0], yy[0], xx[1]-xx[0], yy[1]-yy[0],  # x1,y1 -> dx,dy
              angles='xy', units='xy', scale=1,
              pivot='tail', scale_units=None,  # Defaults
              width=sscale,                    # Shaft width
              headwidth=2*hscale, headlength=3*hscale, headaxislength=2*hscale,  # Head size - scales strangely with width
              zorder=zorder,    # zorder <2 places arrow below grid.
              **kwargs)
    
    return



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Marc van der Sluys' personal Python modules.

The code is being developed by `Marc van der Sluys <http://marc.vandersluys.nl>`_ of the Institute of Nuclear
and High-Energy Physics (Nikhef), and the Department of Physics at Utrecht University, both in The
Netherlands.  The sluyspy package can be used under the conditions of the EUPL 1.2 licence.  These pages
contain the API documentation.  For more information on the Python package, licence and source code, see the
`sluyspy GitHub page <https://github.com/MarcvdSluys/sluyspy>`_.
"""

name = 'sluyspy'
__author__ = 'Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl'




# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Functions to deal with WP weather data in the sluyspy package"""

import numpy as _np
import pandas as _pd
import sluyspy.weather as _swtr


def read_36h_forecast_data(wp_dir, loc, verbosity=1):
    """Read WP 36h forecast files (full day today + latest) and combine them.
    
    Parameters:
      wp_dir (str):     Directory containing the WP data files.
      loc (str):        Name of the town to read data for.
      verbosity (int):  Verbosity (0-4 for silent to loud).
    
    Returns:
      (pd.df):  Pandas.DataFrame containing time, clouds, rain, temp, press, RH, W.speed, W.dir.
    """
    
    import datetime as dt
    today = dt.date.today()
    
    # Set WP input file name:
    WP36File = wp_dir+'wp_weer_'+today.strftime('%Y-%m-%d')+'_36h.dat'
    
    # Read today's forecast:
    if verbosity > 1: print('- '+WP36File)
    df_today    = read_36h_forecast_file(WP36File, loc, verbosity)
    
    # Read tomorrow's forecast:
    if verbosity > 1: print('- '+wp_dir+'wp_weer_latest_36h.dat')
    df_tomorrow = read_36h_forecast_file(wp_dir+'wp_weer_latest_36h.dat', loc, verbosity)
    
    # Combine the two datasets:
    if (df_today is None) and (df_tomorrow is None):
        return None
    elif df_today is None:
        df_combined = df_tomorrow
    elif df_tomorrow is None:
        df_combined = df_today
    else:
        df_combined  = df_tomorrow.combine_first(df_today)
    
    if verbosity > 3: print('Combined WP data:\n', df_combined)
    
    return df_combined


def read_36h_forecast_file(file_name, loc, verbosity=1):
    """Read the forecast for the given location from a single WP 36h data file.
    
    Parameters:
      file_name (str):  Name of the input file.
      loc (str):        Location to read data for.
      verbosity (int):  Verbosity (0-4 for silent to loud).
    
    Returns:
      (pandas.DataFrame):  Forecast data for the location provided.  None if no data were found.
    """
    
    in_file = open(file_name,'r')
    
    # Determine 'header', ie number of lines to skip, by reading every line and checking for the location name:
    n_header=0
    while n_header > -1:
        n_header += 1
        line = in_file.readline()
        
        if loc in line: break
        if n_header > 442:
            return None
        
    n_header += 2  # Skip the two header lines
    
    wp_data = _np.genfromtxt(file_name, skip_header=n_header, max_rows=36)  # Read the data
    
    
    # Convert Numpy array to Pandas dataframe:
    cols         = ['year','month','day','time', 'clouds', 'rain', 'temp', 'press', 'rh', 'ws', 'wd']
    time0        = wp_data[0,3]
    idx          = range(36)+time0  # Set the index to time in hours since midnight today
    wp_data[:,3] = idx              # Set time in hours since midnight today
    df           = _pd.DataFrame(data=wp_data, index=idx, columns=cols)
    
    # Add datetime column 'dtm' to df:
    df['hour'] = df.time % 24  # Clock time in hours (0-23)
    df['dtm'] = _pd.to_datetime(df[['year','month','day','hour']])  # Turn the columns in the df into a single datetime column
    del df['year'],df['month'],df['day'],df['hour']  # No longer needed
    df = df[['dtm'] + [x for x in df.columns if x != 'dtm']]  # Move datetime column to front
    
    # Convert "wind speed" (actually force) to proper wind speed in m/s:
    df.ws = 0.836 * _np.power(df.ws, 3/2)  # v = 0.836 B^(3/2) m/s - https://en.wikipedia.org/wiki/Beaufort_scale#History
    
    # Compute derived variables:
    df['wchil'] = _swtr.wind_chill_temperature(df.temp, df.ws)              # Wind chill (deg C)
    df['dp'] = _swtr.dew_point_from_tempc_rh(df.temp, df.rh/100)            # Dew point (deg C)
    df['ah'] = _swtr.absolute_humidity_from_tempc_rh(df.temp, df.rh/100)    # Absolute humidity (g/m^3)
    
    if verbosity > 4: print('Single-file WP data:\n', df)
    
    return df


def smoothen_36h_forecast_data(wpfc, verbosity=1):
    """Smoothen WP 36h forecast data.
    
    Parameters:
      wpfc (pd.df):     pandas.DataFrame containing WP forecast data.
      verbosity (int):  Verbosity (0-4 for silent to loud).
    
    Returns:
      (tuple):  tuple of three pd.dfs (wpfc,wpfci):
      
      - (pd.df): Pandas DataFrame containing WP forecast data.
      - (pd.df): Pandas DataFrame containing interpolated WP rain forecast data.
    """
    
    if verbosity > 0: print('Smoothening WP forecast...')
    if wpfc is None: return None,None  # Issue with WP data
    
    from scipy import interpolate as _ipol
    
    
    # ### WIND SPEED ###
    # Smoothen the predicted WP wind speed (coarse because it was force) with a fit:
    wpfc['ws_fit'] = None
    if wpfc.ws is not None:
        coefficients = _np.polyfit(wpfc.time, wpfc.ws, 11)   # 11-th degree fit
        polynomial = _np.poly1d(coefficients)                # polynomial contains the fit equation (try print(polynomial))
        wpfc.ws_fit = _np.maximum(polynomial(wpfc.time),0)   # Compute fit points for x; Wind speed shouldn't be negative
        
        wpfc.ws = wpfc.ws.round(1)
        
    # Use a spline interpolation to fit the WP predicted wind 'speed' instead:
    # spl_coefs = _ipol.splrep(wpfc.time, wpfc.ws)         # Compute the spline coefficients
    # wpfci['time'] = _np.arange(0,360)/10
    # wpfci['ws'] = _ipol.splev(wpfci['time'], spl_coefs)  # Do the spline interpolation
    # wpfci.ws[wpfci.ws<0] = 0                                   # Wind speed shouldn't be negative
    
    # Compute derived variables:
    wpfc['wchil'] = _swtr.wind_chill_temperature(wpfc.temp, wpfc.ws_fit)  # Wind chill (from smoothened wind speed)
    
    
    # ### RAIN ###
    # Use a spline interpolation for the predicted WP rain:
    wpfci = _pd.DataFrame()
    if wpfc is not None:
        spl_coefs = _ipol.splrep(wpfc.time, wpfc.rain)      # Compute the spline coefficients
        ipl_range = min(wpfc.time.iloc[-1]+1, 49)                 # 37-49 -> 36-48h
        wpfci['time'] = _np.arange(0,ipl_range*10+1)/10
        wpfci['rain'] = _ipol.splev(wpfci.time, spl_coefs)  # Do the spline interpolation
        wpfci.rain[wpfci.rain<0] = 0                              # Rain shouldn't be negative
    
    if verbosity > 2:
        print('Raw WP data:\n', wpfc)
        if verbosity > 3: print('Interpolated WP rain data:\n', wpfci)
    
    return wpfc, wpfci




# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Text functions for the sluyspy package"""


def wrap_line(line, wlen, indent=2):
    """Wrap a line at the last space before a specified length.
    
    NOTE: this goes WRONG for if words longer than wlen are present.
    
    Parameters:
      line (str):    Line to wrap.
      wlen (int):    Maximum line length to wrap at.
      indent (int):  Indentation of the continuation line
    
    Returns:
      (str):  The line with extra newline characters.
    """
    
    llen = len(line)
    
    # No need to break:
    if wlen >= llen: return line
    
    i0 = 0     # Base position
    il = wlen  # Move between i0 and i0+wlen
    dl = -1    # Move backwards by default
    wraps = 0
    
    while True:
        # print(i0+il,llen)
        if line[i0+il]==' ':
            wraps += 1
            line = line[0:i0+il]+'\n'+' '*indent+line[i0+il+1:]  # Remove the space and add an indentation of <indent> spaces
            i0 = i0+il+indent
            il = wlen
            
            # print(wraps,wlen,llen,wraps*wlen,len(line),il,i0+il, line)
            if i0+il >= len(line): break
            dl = -1  # Search backward by default
            continue
        
        il += dl
        if i0+il > llen: break
        
        # print('i0,il,i0+il,dl:" ', i0,il,i0+il,dl)
        if il==indent-1:
            i0 += wlen
            dl = +1  # Start searching forward
            continue
        
    return line



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Pandas functions for the sluyspy package"""

import re as _re


def csv_formatted_from_df(df, file_name, fmt):
    """Write a Pandas DataFrame as a formatted csv file.
    
    Parameters:
      df (pd.df):       Pandas DataFrame to write to csv file.
      file_name (str):  Path to the output csv file (may be a Path).
      fmt (str):        C-style formatting string for values w/o newline; the header string will be adapted from this.
    """
    
    fmt += '\n'     # Add newline to value format string
    
    # Construct header format string from value format string:
    hdr_fmt = fmt
    hdr_fmt = _re.sub(r'\.[0-9]*[ifegIFEG]', r's', hdr_fmt)  # Dot + number(s) + ifeg/IFEG
    hdr_fmt = _re.sub(r'\.[0-9]*lf', r's', hdr_fmt)          # Dot + number(s) + lf/LF
    hdr_fmt = _re.sub(r'lf', r's', hdr_fmt)                  # lf
    hdr_fmt = _re.sub(r'LF', r's', hdr_fmt)                  # LF
    hdr_fmt = _re.sub(r'[dicfegDICFEG]', r's', hdr_fmt)      # Any remaining letters w/o numbers
    hdr_fmt = _re.sub(r'%0*', r'%', hdr_fmt)                 # Leading zeros
    
    outfile = open(file_name,'w')
    outfile.write(hdr_fmt % tuple(df.columns.values))  # Write column names as header
    
    for idx,ser in df.iterrows():  # Makes (index, Series) pairs out of rows
        outfile.write(fmt % tuple(ser.values))  # Write formatted row - needs '\n' for EoL
    outfile.close()
    
    return



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Functions for the sluyspy package to do with the Earth"""


import numpy as _np
import astroconst as _ac
import sluyspy.numerics as _snum


def distance(lon1,lat1, lon2,lat2, miles=False):
    """Compute the distance between two points over the Earth's surface.
    
    Parameters:
      lon1 (float):  Longitude of first location (rad).
      lat1 (float):  Latitude of first location (rad).
      lon2 (float):  Longitude of second location (rad).
      lat2 (float):  Latitude of second location (rad).
      miles (bool):  Return result in miles rather than kilometres.  Optional; defaults to False.
    
    Returns:
      (float):  Distance in kilometres (or miles, if desired).
    """
    
    r_e  = _ac.earth_r*1e-3                # Earth's radius in km
    fl   = 0.003352810665                  # Earth's flattening
    
    mlat  = (lat1+lat2)/2
    dlat2 = (lat1-lat2)/2
    dlon2 = (lon1-lon2)/2
    
    sins = _np.sin(dlat2)**2 * _np.cos(dlon2)**2 + _np.cos(mlat)**2 * _np.sin(dlon2)**2
    coss = _np.cos(dlat2)**2 * _np.cos(dlon2)**2 + _np.sin(mlat)**2 * _np.sin(dlon2)**2
    rat  = _np.arctan2(_np.sqrt(sins), _np.sqrt(coss))
    
    r = _np.sqrt(sins*coss)/(rat + _snum.tiny)  # Prevent division by zero - good idea?
    dist = 2 * r_e * rat
    
    h1 = (3*r-1) / (2*coss + _snum.tiny)
    h2 = (3*r+1) / (2*sins + _snum.tiny)
    
    distance = dist*(1  +  fl*h1*_np.sin(mlat)**2 * _np.cos(dlat2)**2  -  fl*h2*_np.cos(mlat)**2 * _np.sin(dlat2)**2)
    if miles: distance = distance * 0.62137119  # Miles rather than km - this is just one of the many definitions of a mile!
    
    return distance


def effective_radius_from_latitude(lat):
    """Compute the effective radius of the Earth from the latitude.
    
    Parameters:
      lat (float):  geographical latitude (radians).
    
    Returns:
      (float):  effective radius of the Earth (metres).
    """
    
    abslat = _np.abs(lat)
    
    Reff = 2 * 9.780356*(1+0.0052885*_np.square(_np.sin(abslat)) - 0.0000059*_np.square(_np.sin(2*abslat))) \
        / (3.085462e-6 + 2.27e-9 * _np.cos(2*abslat) - 2e-12 * _np.cos(4*abslat))  # Earth radius in metres
    
    return Reff


def distance_direction_from_longitude_latitude(lon, lat):
    """Compute the distance and direction between points for two arrays of longitude and latitude coordinates.
    
    The distance is in kilometres, the direction angle in radians, where 0=N, pi/2=E, +/-pi=S and -pi/2=W.
    
    Parameters:
      lon (float):  array of (at least two) longitude coordinates (radians).
      lat (float):  array of (at least two) latitude coordinates (radians).
    
    Returns:
      (tuple):  (ddist, direc):
    
      ddist (float):  array with distances between the subsequent coordinate sets (km).  ddist[0]=NaN.
      direc (float):  array with direction angles between the subsequent coordinate sets (radians).  ddist[0]=NaN.

    """
    
    Re     = effective_radius_from_latitude(lat)/1000  # Lat in radians, Re in m -> km
    rad2km = Re  # 2 pi rad = 2 pi Re km
    xpos   = lon * _np.cos(lat) * rad2km
    ypos   = lat * rad2km
    dxpos  = xpos - xpos.shift(1)
    dypos  = ypos - ypos.shift(1)
    ddist  = _np.sqrt(_np.square(dxpos) + _np.square(dypos))  # km
    direc  = _np.arctan2(dxpos, dypos)  # rad: N=0, E=pi/2, S=+/-pi, W=-pi/2 rad
    
    return ddist, direc



#!/bin/env python
# PYTHON_ARGCOMPLETE_OK
# -*- coding: utf-8 -*-

""" test-to_precision.py:  Demonstrate the to_precision code.
    Eric Moyer github.com/epmoyer eric@lemoncrab.com
"""


import colored_traceback
colored_traceback.add_hook()


__author__ = 'Eric Moyer github.com/epmoyer eric@lemoncrab.com'


from math import log10
from tabulate import tabulate
from sluyspy.to_precision import to_precision
import sluyspy.numerics as snum

print(to_precision(0.075, 1))
print(snum.sigdig(0.075, 1))


def main():
    """
    Displays a table containing example conversions for to_precision()
    """
    
    # Build test values
    seed = [float(int(123456789. / 10**x)) for x in range(7, -1, -1)]
    test_values = ([0.0, 1.0, 10.0, 100.0, -1.0] +
                   [x for x in seed] +
                   [x / 10**int(log10(x)) for x in seed] +
                   [x / 10**9  for x in seed])
    
    option_cases = (
        ('Default (Auto Notation)',                     dict()),
        ('Standard Notation',                           dict(notation='std')),
        ('Scientific Notation',                         dict(notation='sci')),
        ('Standard Notation with zero stripping',       dict(notation='std', strip_zeros=True)),
        ('Scientific Notation with zero stripping',     dict(notation='sci', strip_zeros=True)),
        ('Standard Notation with integer preservation', dict(notation='std', preserve_int=True)),
        ('Auto Notation with exponent limit of 5',      dict(auto_limit=5)),
    )
    
    precisions = tuple(range(1, 6))
    
    # prints out the label, function call, and precision table
    for options_description, options_dict in option_cases:
        
        '''
        Prints label for table.
        Ex:
        Default (Auto Notation):
            to_precision(value, precision)
        '''
        print(options_description + ':')
        options_string = ', '.join(
            ['value', 'precision'] +
            [note + '=' + repr(inputs) for note, inputs in options_dict.items()])
        print('to_precision({inputs})'.format(inputs=options_string), end='\n' * 2)
        
        table = []
        for val in test_values:
            table_row = ['{:0.10f}'.format(val).rstrip('0').rstrip('.')]
            for precision in precisions:
                result_string = to_precision(val, precision, **options_dict)
                table_row.append(result_string)
            table.append(table_row)
        
        headers = ['value'] + ['precision={}'.format(x) for x in precisions]
        
        print(tabulate(table, headers, disable_numparse=True), end='\n' * 3)
    return


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Fitting functions for the sluyspy package"""


import numpy as _np
import pandas.core as _pdc
import sluyspy.cli as _scli


def np_polyfit_chi2(xvals, yvals, order, ysigmas=None, verbosity=0):
    """Wrapper for NumPy polyfit, that returns the fitting coefficients and reduced chi squared.
       Print details if desired.
    
    Note: does not compute the uncertainties in the coefficients; use scipy_curvefit_chi2() for that.
    
    Parameters:
      xvals (float):    X values.
      yvals (float):    Y values.
      order (int):      Polynomial order for fit.
      ysigmas (float):  Uncertainties (errors, standard deviations) in the y values; defaults to None -> all sigmas=1.
      verbosity (int):  Verbosity:  0: quiet,  1: print red.chi2,  2: print more,  3: print table,  4: print all;  defaults to 0.
    
    Returns:
      (tuple):  Tuple consisting of (coefs, red_chi2):
    
      - coefs (float):     Array of fitting coefficients, with length (order+1).
      - red_chi2 (float):  Reduced chi squared (chi^2 / (n-m)) for the fit.
    """
    
    if ysigmas is None:
        if verbosity>0: print('ysigmas=None; assuming sigma=1 for all data points.')
        coefs, resids, rank, sing_vals, rcond = _np.polyfit(xvals, yvals, order, full=True)
        ysigmas = yvals*0 + 1  # This is implicitly assumed in the previous line; works for pd.Series and numpy arrays
    else:
        coefs, resids, rank, sing_vals, rcond = _np.polyfit(xvals, yvals, order, w=1/ysigmas, full=True)  # Note, not 1/sigma**2!
    
    
    # Print polyfit-specific details:
    if verbosity>3:
        print('coefs:      ', coefs)
        print('resids:     ', resids)
        print('rank:       ', rank)
        print('sing_vals:  ', sing_vals)
        print('rcond:      ', rcond)
    
    # Compute reduced chi^2 and print general fit details:
    red_chi2 = print_fit_details(xvals,yvals, ysigmas=ysigmas, fittype='np_polyfit', coefs=coefs,
                                 verbosity=verbosity)
    
    return coefs, red_chi2



def scipy_curvefit_chi2(fit_fun, xvals, yvals, coefs0, ysigmas=None, verbosity=0):
    """Wrapper for SciPy's curve_fit, that returns the fitting coefficients and reduced chi^2.
       Print details if desired.
    
    Parameters:
      fit_fun (function):  Fitting function, should look like  def fit_fun(x, a,b,c):  and return y=f(x, a,b,c).
      xvals (float):       Array containing x values to fit.
      yvals (float):       Array containing y values to fit.
      coefs0 (float):      Array with initial guess for fitting coefficients.
      ysigmas (float):     Array containing y sigmas/uncertainties.
      verbosity (int):     Verbosity to stdout (0-4).
    
    Returns:
      (tuple):  Tuple containing (coefs, dcoefs, red_chi2, var_cov, ier):
    
      - coefs (float):     Array containing the final fitting coefficients.
      - dcoefs (float):    Array containing uncertainties on the fitting coefficients.
      - red_chi2 (float):  Reduced chi squared (chi2/(n-m)).
      - var_cov (float):   2D array containing the variance-covariance matrix.
      - ier (int):         Return value, >0 if the fit succeeded.
    """
    
    from scipy.optimize import curve_fit as _curve_fit
    
    try:
        # Do the fit:
        coefs, var_cov, infodict, mesg, ier = _curve_fit(fit_fun, xvals, yvals, p0=coefs0, sigma=ysigmas,
                                                         method='lm', full_output=True)
    
    # If call failed:
    except Exception as e:
        if verbosity>0:
            print(__name__+': fit failed: ', end='')
            print(e)
        ier = -1
    
    # If call failed or fit did not converge:
    if ier<=0:
        if verbosity>0: print('Fit failed: %i' % ier)
        return None, None, None, None, ier
    
    
    # Print specific curve_fit output:
    if verbosity>3:
        print('Fit success: ', ier)
        print('Message: ', mesg)
        if verbosity>4:
            print('Info dict:\n', infodict, '\n')
        print('Initial coefficients:  ', coefs0)
        print('Final coefficients:    ', coefs)
        print('Variance/covariance matrix: \n', var_cov)
    
    
    # Compute some fit-quality parameters:
    if ysigmas is None: ysigmas = yvals*0 + 1  # This is implicitly assumed in the previous line; works for pd.Series and numpy arrays
    
    dcoefs = _np.sqrt(_np.diag(var_cov))              # Standard deviations on the coefficients
    
    # Compute reduced chi^2 and print general fit details:
    red_chi2 = print_fit_details(xvals,yvals,ysigmas, fittype='scipy_curvefit', coefs=coefs, dcoefs=dcoefs,
                                 var_cov=var_cov, fit_fun=fit_fun, verbosity=verbosity)
    
    return coefs, dcoefs, red_chi2, var_cov, ier


def print_fit_details(xvals,yvals, ysigmas=None, fittype=None, coefs=None,dcoefs=None, var_cov=None,
                      fit_fun=None, yfit=None, verbosity=2, abs_diff=True,rel_diff=True, sigdig=6,
                      coef_names=None, coef_facs=None, rev_coefs=None, yprintfac=1, rel_as_pct=False):
    
    """Compute and return the reduced chi^2, and print other fit details if desired.
    
    Calculations are done without fitting, so that this function can be called after a fit.  The fit values
    will be computed from coefs, xvals using fit_fun if specified.  In fact, two series of yvals ('true' and
    'fit') can be specified without any fitting ever, by setting fittype=coefs=xvals=None and specifying yfit.
    This will print statistics on the comparison of the two data sets, similar to those after a fit.  If xvals
    is specified, a bit more detail can be printed.
    
    Parameters:
      xvals (float):      Array with x values.
      yvals (float):      Array with y values.
    
      ysigmas (float):    Array with y sigmas; optional, defaults to None (i.e. using 1 for all ysigmas).
    
      fittype (str):      Type of fit: 'np_polyfit', 'scipy_curvefit'; optional, defaults to None.
      coefs (float):      Array with fit coefficients; optional, defaults to None.
    
      dcoefs (float):     Uncertainties in coefficients from scipy_curvefit; optional, defaults to None.
      var_cov (float):    Variance-covariance matrix in 2D NumPy array; optional, defaults to None.
      fit_fun (fun):      Fit function used for scipy_curvefit; optional, defaults to None.
      yfit (float):       "Fit" values for Y for fittype=None; optional, defaults to None.
    
      verbosity (int):    Verbosity (0-3); optional - defaults to 2.
      abs_diff (bool):    Print absolute differences (optional; defaults to True).
      rel_diff (bool):    Print relative differences (optional; defaults to True).
      sigdig (int):       Number of significant digits to print (optional, defaults to 6).
    
      coef_names (str):   Array of coefficient names for printing (optional, defaults to None).
      coef_facs (float):  Array of coefficient multiplication factors for printing (optional; defaults to None,
                          in which case 1 is used).  Useful for e.g. printing degrees when radians are fitted.
      rev_coefs (bool):   Reverse printing order in coefficient table: last coefficient at top.
                          Optional, defaults to None, resulting in True for fittype np_polyfit, False otherwise.
    
      yprintfac (float):  Multiplication factor when printing absolute differences in y (optional, defaults to 1).
      rel_as_pct (bool):  Print relative difference as percentage rather than fraction (optional, defaults to False).
    
    Returns:
      (float):            Reduced Chi^2.
    
    Note:
      - If fittype = 'scipy_curvefit', fit_fun must be specified.
      - If fittype is None, yfit values must be specified, since they will not be computed from xvals
        and coefs.  In that case, coefs=xvals=None can be specified.
    """
    
    from sluyspy.numerics import sigdig as sd
    
    if fittype=='np_polyfit':
        yfit      = _np.poly1d(coefs)(xvals)
        if rev_coefs is None: rev_coefs = True
    elif fittype=='scipy_curvefit':
        if fit_fun is None: _scli.error('A fit function must be specified for fittype '+str(fittype))
        if coefs is None: _scli.error('Fit coefficients must be specified for fittype '+str(fittype))
        yfit      = fit_fun(xvals, *coefs)
        if rev_coefs is None: rev_coefs = False
    elif fittype is None:
        if yfit is None: _scli.error('yfit values must be specified for fittype '+str(fittype))
        if rev_coefs is None: rev_coefs = False
    else:
        _scli.error('Unknown fittype: '+fittype+'; must be one of "np_polyfit", "scipy_curvefit" or None.')
    
    yfit = yvals*0 + yfit  # Ensure yfit has same type as xvals (np.array/pd.Series)
    
    if ysigmas is None:
        if verbosity>1: print('\nysigmas=None; assuming sigma=1 for all data points.\n')
        ysigmas = yvals*0 + 1  # Set ysigmas to 1 - works for pd.Series and np.arrays
    else:
        # Discard data points with sigma == 0:
        ndat = len(ysigmas)                        # Number of data points
        if xvals is not None:  xvals = xvals[ysigmas>0]
        if yfit is not None:   yfit  = yfit[ysigmas>0]
        yvals   = yvals[ysigmas>0]
        ysigmas = ysigmas[ysigmas>0]
        if (verbosity>1) and (len(ysigmas) != ndat): print('\nDiscarding data points with ysigma=0.\n')
    
    if verbosity>1: mean_ysigma = _np.nanmean(ysigmas) * yprintfac  # Mean sigma in y
    
    
    rel_fac = 1  # Print fraction
    rel_str = ' '
    if rel_as_pct:
        rel_fac = 100  # Print percentage
        rel_str = '%'
        
    
    # Compute reduced chi^2:
    if coefs is None:
        ncoefs = 0                                          # Number of coefficients
    else:
        ncoefs = len(coefs)                                 # Number of coefficients
    
    ndat                = len(yvals)                        # Number of data points
    ydiffs              = yfit - yvals                      # Differences/residuals
    yresids             = ydiffs/ysigmas                    # Weighted residuals
    chi2                = _np.sum(yresids**2)               # Chi^2 - NumPy needed for large numbers
    red_chi2            = chi2/(ndat-ncoefs)                # Reduced chi^2
    
    abs_ydiffs          = ydiffs                            # Absolute differences in y
    mean_abs_ydiff      = _np.nanmean(abs_ydiffs)           # The mean absolute difference between data and fit values
    med_abs_ydiff       = _np.nanmedian(abs_ydiffs)         # The median absolute difference between data and fit values
    
    abs_abs_ydiffs      = _np.abs(ydiffs)                   # Absolute values of absolute differences in y
    mean_abs_abs_ydiff  = _np.nanmean(abs_abs_ydiffs)       # The mean absolute difference between data and fit values
    med_abs_abs_ydiff   = _np.nanmedian(abs_abs_ydiffs)     # The median absolute difference between data and fit values
    
    rel_ydiffs          = ydiffs[yvals!=0]/yvals[yvals!=0]  # Relative differences in y
    rel_ydiffs         *= rel_fac                           # Fraction or percentage?
    mean_rel_ydiff      = _np.nanmean(rel_ydiffs)           # The mean absolute difference between data and fit values
    med_rel_ydiff       = _np.nanmedian(rel_ydiffs)         # The median absolute difference between data and fit values
    
    abs_rel_ydiffs      = _np.abs(rel_ydiffs)               # Absolute values of relative differences in y
    mean_abs_rel_ydiff  = _np.nanmean(abs_rel_ydiffs)       # The mean absolute difference between data and fit values
    med_abs_rel_ydiff   = _np.nanmedian(abs_rel_ydiffs)     # The median absolute difference between data and fit values
    
    
    # Compute and print details:
    if verbosity>0:
        
        # Find maximum differences:
        if verbosity>2:
            max_abs_diff_y  = _np.max(abs_abs_ydiffs)  # Maximum absolute difference in y
            max_rel_diff_y  = _np.max(abs_rel_ydiffs)  # Maximum relative difference in y
            
            if type(xvals) is _pdc.series.Series:
                max_abs_diff_x  = xvals[abs_abs_ydiffs == max_abs_diff_y].to_numpy()[0]  # x value for maximum absolute difference (Pandas)
                max_rel_diff_x  = xvals[ _np.logical_and(yvals!=0, abs_rel_ydiffs == max_rel_diff_y)].to_numpy()[0]  # x value for maximum relative difference (Pandas)
            elif xvals is None:
                max_abs_diff_x  = None
                max_rel_diff_x  = None
                xvals = yvals*0 + _np.nan  # Fill with NaNs
            else:
                max_abs_diff_x  = xvals[abs_abs_ydiffs == max_abs_diff_y][0]             # x value for maximum absolute difference (Numpy)
                max_rel_diff_x  = xvals[ _np.logical_and(yvals!=0, abs_rel_ydiffs == max_rel_diff_y)][0]  # x value for maximum relative difference (Numpy)
        
        # Print details:
        if verbosity>1:
            print('Fit type:      ', fittype)
            if fit_fun is not None:  print('Fit function:  ', fit_fun.__name__)
            print()
            
            print('Fit quality:')
            print('Number of data points:            ', ndat)
            if verbosity>2:
                if ncoefs>0:
                    print('Number of coefficients:           ', ncoefs)
                    print('Degrees of freedom:               ', ndat - ncoefs)
                    print()
                print('Chi2:                             ', sd(chi2, sigdig))
            
        print('Reduced chi2:                     ', sd(red_chi2, sigdig))
        if verbosity>1:
            print('sqrt reduced chi2:                ', sd(_np.sqrt(red_chi2), sigdig))
            if mean_ysigma != 1: print('sqrt red.chi2 * mean sigma:       ', sd(_np.sqrt(red_chi2)*mean_ysigma, sigdig))
        
        if abs_diff:
            if verbosity>1: print()
            print('Mean/med. |absolute difference|:  ', sd(mean_abs_abs_ydiff*yprintfac, sigdig), '  /  ',
                  sd(med_abs_abs_ydiff*yprintfac, sigdig))
            
            if verbosity>2:
                print('Max. |absolute difference|:       ', sd(max_abs_diff_y*yprintfac, sigdig),
                      '  @   x =', sd(max_abs_diff_x, sigdig))
                
                print('Mean/med. absolute difference:    ', sd(mean_abs_ydiff*yprintfac, sigdig), '  /  ',
                      sd(med_abs_ydiff*yprintfac, sigdig))
            
        if rel_diff:
            if verbosity>2: print()
            print('Mean/med. |relative difference|:  ', sd(mean_abs_rel_ydiff, sigdig), rel_str, '  /  ',
                  sd(med_abs_rel_ydiff, sigdig), rel_str)
            
            if verbosity>2:
                print('Max. |relative difference|:       ', sd(max_rel_diff_y, sigdig), rel_str,
                      '  @   x =', sd(max_rel_diff_x, sigdig))
                
                print('Mean/med. relative difference:    ', sd(mean_rel_ydiff, sigdig), rel_str, '  /  ',
                      sd(med_rel_ydiff, sigdig), rel_str)
            
        
        # Print fit coefficients:
        if (verbosity>1) and (coefs is not None):
            if coef_facs is None:  coef_facs = _np.array(coefs)*0+1  # Coefficient print factor is 1 by default:
            
            # Give all coefficient names the same length:
            if coef_names is not None:
                strlen = len(max(coef_names, key=len))  # Length of the longest string in the list
                fmt = ' %'+str(strlen)+'s'
                for icoef in range(ncoefs):
                    coef_names[icoef] = fmt % (coef_names[icoef])
            
            print('\nFit coefficients', end='')
            if rev_coefs: print(' (reversed)', end='')
            print(':')
            for icoef in range(ncoefs):
                jcoef = icoef
                if rev_coefs: jcoef = ncoefs-icoef-1
                
                print(' c%1i:' % (icoef), end='')  # Nr
                
                if coef_names is not None:
                    print(coef_names[jcoef]+': ', end='')  # Name
                
                print(' %12.5e' % (coefs[jcoef] * coef_facs[jcoef]), end='')  # Value
                
                if dcoefs is not None:
                    print(' ± %12.5e (%9.2f%%)' % (dcoefs[jcoef] * coef_facs[jcoef],
                                                   _np.abs(dcoefs[jcoef]/coefs[jcoef]*100) ), end='')
                
                print()
        
        # Print correlation and variance-covariance matrices:
        if (verbosity>2) and (var_cov is not None):
            print('\nCorrelation matrix:')
            corr = correlation_matrix_from_variance_covariance_matrix(var_cov)
            _print_matrix(corr, 'corr', coef_names)
            
            if verbosity>3:
                print('\nVariance-covariance matrix:')
                _print_matrix(var_cov, 'var_cov', coef_names)
        
        # Print all fit data points:
        if verbosity>4:
            print('\nFit data:')
            print('%9s  %12s  %12s  %12s  %12s  %12s  %12s  %12s' %
                  ('i', 'x_val', 'y_val', 'y_sigma', 'y_fit', 'y_diff_abs', 'y_diff_wgt', '|y_dif_rel|') )
            
            if type(yvals) is _pdc.series.Series:
                for ival in range(ndat):
                    print('%9i  %12.5e  %12.5e  %12.5e  %12.5e  %12.5e  %12.5e  %12.5e' %
                          (ival, xvals.iloc[ival],yvals.iloc[ival]*yprintfac, ysigmas.iloc[ival]*yprintfac,
                           yfit.iloc[ival]*yprintfac, ydiffs.iloc[ival]*yprintfac,
                           yresids.iloc[ival]*yprintfac, _np.abs(ydiffs.iloc[ival]/yvals.iloc[ival])*rel_fac ) )
            
            else:  # Probably Numpy:
                for ival in range(ndat):
                    print('%9i  %12.5e  %12.5e  %12.5e  %12.5e  %12.5e  %12.5e  %12.5e' %
                          (ival, xvals[ival],yvals[ival]*yprintfac, ysigmas[ival]*yprintfac,
                           yfit[ival]*yprintfac, ydiffs[ival]*yprintfac, yresids[ival]*yprintfac,
                           _np.abs(ydiffs[ival]/yvals[ival])*rel_fac ) )
            
            print('%9s  %12s  %12s  %12s  %12s  %12s  %12s  %12s' %
                  ('i', 'x_val', 'y_val', 'y_sigma', 'y_fit', 'y_diff_abs', 'y_diff_wgt', '|y_dif_rel|') )
            
    if verbosity>1: print()
    
    return red_chi2


def correlation_matrix_from_variance_covariance_matrix(var_cov):
    """Compute the normalised correlation matrix from a variance-covariance matrix.
    
    Parameters:
      var_cov (float):  2D NumPy array containing the variance-covariance matrix.
    
    Returns:
      (float): 2D NumPy array containing the normalised-correlation matrix.
    """
    
    corr    = _np.zeros_like(var_cov)
    matsize = _np.shape(corr)[0]
    
    for i in range(matsize):
        for j in range(matsize):
            corr[i][j] = var_cov[i,j] / _np.sqrt(var_cov[i,i] * var_cov[j,j])
    
    return corr


def polynomial(x, *coefs):
    """Return a polynomial y = Sum_i=0^N-1 coef_i x^i.
    
    Parameters:
      x (float):      X value(s).
      coefs (float):  Array of polynomial coefficients.
                      The polynomial order is defined by the number of coefficients in the array.
    
    Returns:
      (float):  Polynomial value.
    """
    
    y = 0
    i = 0
    for coef in coefs:
        y += coef * _np.power(x, i)
        i += 1
    
    return y


def _print_matrix(mat, mat_type, coef_names=None):
    """Print the contents of a 2D (correlation or covariance) matrix.
    
    Parameters:
      mat (float):  2D NumPy array containing the variance-covariance matrix.
    """
    
    # Matrix size:
    matsize = _np.shape(mat)[0]
    
    # Derive fomatting for names and numbers from matrix type:
    if mat_type == 'corr':
        namfmt = ' %7s'
        maxnamlen = 7
        numfmt = '%8.4f'   # ' -0.1238'
    else:
        namfmt = ' %12s'
        maxnamlen = 12
        numfmt = '%13.5e'  # ' -1.12345e-13'
        
    # Length of coefficient names:
    if coef_names is None:
        namlen = 0
    else:
        namlen = len(coef_names[0])
    
    # Print row of coefficient numbers:
    print(' '*(3+namlen), end='')  # Spaces: c_i + name length
    for i in range(matsize):
        print(namfmt % ('c'+str(i)), end='')
    print()
        
    # Print row of coefficient names:
    if coef_names is not None:
        print(' '*(3+namlen), end='')  # Spaces: c_i + name length
        for i in range(matsize):
            print(namfmt % (coef_names[i].strip()[:maxnamlen]), end='')  # Cut off name to preserve table formatting
        print()
    
    # Print columns with coefficient numbers, coefficient names and values:
    for i in range(matsize):
        if matsize > 10:
            print('c%2i' % (i), end='')
        else:
            print('c%1i ' % (i), end='')
            
        if coef_names is not None:  print(coef_names[i], end='')
        for j in range(matsize):
            print(numfmt % (mat[i,j]), end='')
        print()
        
    return




# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#  
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#  
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""HTML functions for the sluyspy package."""

# from .cli import error as _error
import datetime as _dt


def start_html_file(file_name='index.html', lang='en', title='Page title', icon=None, css=None,
                    author='Marc van der Sluys', copyr_start=None, refresh=None, meta_prop=None):
    
    """Create an html file, write the head section and start the body.
    
    Parameters:
      file_name (str):    Name/path of the html file.
      lang (str):         Language.
      title (str):        Page title.
      icon (str):         Path to an icon file.
      css (str):          Path to a css file.
      author (str):       Author name.
      copyr_start (int):  Starting year of copyright.
      refresh (int):      Refresh period in minutes.
      meta_prop (dict):   Dict of meta data with the keys containing properties and the values
                          their content.
    
    Returns:
      (io):  File descriptor.
    """
    
    # Create the HTML file:
    f = open(file_name, 'w')
    
    f.write('<!DOCTYPE HTML>\n')
    f.write('<html lang="'+lang+'">\n')
    
    # Write the <head> section:
    f.write('  <head>\n')
    f.write('    <meta http-equiv="Content-Type" content="text/html;charset=utf-8">\n')
    
    # Add refresh data if desired:
    if refresh is not None:
        f.write('    <meta http-equiv="Refresh" content="'+str(refresh*60)+'">\n')
        
    # Add an icon or css if desired:
    if icon is not None:    f.write('    <link rel="icon" href="'+icon+'">\n')
    if css is not None:     f.write('    <link rel="stylesheet" type="text/css" href="'+css+'">\n')
    
    # Add a title:
    f.write('    <title>'+title+'</title>\n')
    
    # Add an author and copyright if desired:
    if author != '':
        current_year = _dt.date.today().year
        if (copyr_start is None) or (copyr_start == current_year):
            f.write('    <meta name="author" content="(c) '+str(current_year)+' '+author+'">\n')
        else:
            f.write('    <meta name="author" content="(c) '+str(copyr_start)+'-'+str(current_year)+' '+author+'">\n')
    
    # Add meta data if any:
    if meta_prop is not None:
        for key,value in meta_prop.items():
            f.write('    <meta property="'+key+'" content="'+value+'">\n')
    
    # Close the <head> section and start the <body> section:
    f.write('  </head>\n')
    f.write('  \n')
    f.write('  <body>\n')
    
    return f


def close_html_file(fd, sc_id=None, sc_secr=None, sc_name=None):
    """Close an html file by writing StatCounter code if desired, and closing <body>, <html> and the file.
    
    Parameters:
      fd (io):        File descriptor.
      sc_id (int):    StatCounter project ID.
      sc_secr (str):  StatCounter security secret.
      sc_name (str):  Name for the StatCounter code block.
    """
    
    # Create some space:
    fd.write('    \n')
    fd.write('    \n')
    
    
    # Write a StatCounter code block if desired:
    if (sc_id is not None) and (sc_secr is not None):
        if sc_name is not None:
            fd.write('    <!-- Start of StatCounter Code for '+sc_name+' -->\n')
        else:
            fd.write('    <!-- Start of StatCounter Code -->\n')
        
        fd.write('    <script type="text/javascript">\n')
        fd.write('    var sc_project='+str(sc_id)+'; \n')
        fd.write('    var sc_invisible=1; \n')
        fd.write('    var sc_security="'+str(sc_secr)+'"; \n')
        fd.write('    var scJsHost = (("https:" == document.location.protocol) ?\n')
        fd.write('    "https://secure." : "http://www.");\n')
        fd.write('    document.write("<sc"+"ript type=''text/javascript'' src=''" +\n')
        fd.write('    scJsHost+\n')
        fd.write('    "statcounter.com/counter/counter.js''></"+"script>");\n')
        fd.write('    </script>\n')
        fd.write('    <noscript><div class="statcounter"><a title="web analytics"\n')
        fd.write('    href="http://statcounter.com/" target="_blank"><img\n')
        fd.write('    class="statcounter"\n')
        fd.write('    src="//c.statcounter.com/'+str(sc_id)+'/0/'+str(sc_secr)+'/1/" alt="web\n')
        fd.write('    analytics"></a></div></noscript>\n')
        if sc_name is not None:
            fd.write('    <!-- End of StatCounter Code for '+sc_name+' -->\n')
        else:
            fd.write('    <!-- End of StatCounter Code -->\n')
        
        fd.write('    \n')
        fd.write('    \n')
    
    
    # Close the <body> and <html> sections:
    fd.write('  </body>\n')
    fd.write('</html>\n')
    
    # Close the file:
    fd.close()
    
    return


def table_td_tr(indent, td_width, td_extra_width):
    """Define trtd, tdtd and tdtr elements for an html table.
    
    Parameters:
      indent (int):          Number of spaces for indentation.
      td_width (str):        Width of empty column between columns (e.g. '1%').
      td_extra_width (str):  Width of extra-wide empty column between columns (e.g. '10%').
    
    Returns:
      (tuple):  tuple containing trtd, tdtd, tdtdw, tdtr:
    
      - trtd (str):    the <tr><td> element.
      - tdtd (str):    the default </td><td> element.
      - tdtdw (str):   the extra-wide </td><td> element.
      - tdbrtd (str):  an empty </td><td> element for an empty row (<br>).
      - tdtr (str):    the </td></tr> element.
    """
    
    trtd    = ' '*indent + '<tr><td>'
    tdtd    = ' '*indent + '</td><td width="' + td_width       + '"></td><td>'
    tdtdw   = ' '*indent + '</td><td width="' + td_extra_width + '"></td><td>'
    tdbrtd  = ' '*indent + '</td><td width="' + td_width       + '"><br></td><td>'
    tdtr    = ' '*indent + '</td></tr>\n'
    
    return trtd, tdtd, tdtdw, tdbrtd, tdtr


def last_update(fd, indent, size='65%', seconds=False, tz=False):
    """Add a 'Last update' line to an html file with the current system date and time.
    
    Parameters:
      fd (io):         File descriptor.
      indent (int):    Number of spaces for indentation.
      size (str):      String with font size, e.g. '100%'.
      seconds (bool):  Print seconds in timestamp.
      tz (bool):       Print time zone in timestamp.
    """
    
    import time
    if seconds:
        time_str = _dt.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    else:
        time_str = _dt.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M')
    
    if tz:
        time_str += ' ' + time.tzname[time.localtime().tm_isdst]  # Add current tz, accounting for DST
    
    fd.write(' '*indent + '<br>\n')
    fd.write(' '*indent + '<p style="font-size:'+size+'; text-align:center; margin:0;">Last update: '+time_str+'</p>\n')
    fd.write(' '*indent + '\n')
    return


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Gravitational-wave functions for the sluyspy package"""


import numpy as _np
import pandas as _pd
import astroconst as _ac


def cbc_waveform(m1,m2, dist,cosi, tlen,tcoal, Npts, risco_fac=1.5, Fplcr=2/5, verbosity=0):
    """Compute a simple, Newtonian(!) compact-binary-coalescence waveform.
    
    Parameters:
      m1 (float):         Mass of binary component 1 (kg).
      m2 (float):         Mass of binary component 2 (kg).
      dist (float):       Distance of binary (m).
      cosi (float):       Cosine of the inclination angle: +/-1 = face on, 0 = edge-on.
      tlen (float):       Length of time of array (s).
      tcoal (float):      Coalescence time (s).
      risco_fac (float):  R_isco = R_star * risco_fac (-, defaults to 1.5); used to compute f_max.
      Npts (float):       Number of data points (-).
    
      Fplcr (float):      The mean value of F_+ and F_x: <F> ~ sqrt(<F_+^2 + F_x^2>),
                          which includes the inclination.  Defaults to 2/5.
      verbosity (int):    Verbosity level, defaults to 0: no output.
    
    Returns:
      (pd.df):  Pandas dataframe containing variables.
    """
    
    mt = m1+m2
    mu = m1*m2/mt
    Mc = (m1*m2/mt**(1/3))**(3/5)
    
    R1 = radius_BH_NS_from_mass(m1)  # Get the BH/NS radius for this mass (m)
    R2 = radius_BH_NS_from_mass(m2)
    a_min = R1 + R2                  # (m)
    
    f_isco = _ac.c**3 / (6**(3/2) * _ac.pi * _ac.g * mt)
    # f_max  = f_isco
    # f_max  = _ac.c**3 / (1**(3/2) * _ac.pi * _ac.g * mt)
    f_max = 1/_ac.pi * _np.sqrt(_ac.G * mt / (risco_fac*a_min)**3)  # f_max based on 1.5 x (sum of radii) and Kepler - same as above, but allows NSs
    
    tstart = tcoal - tlen
    
    # Create DataFrame:
    df = _pd.DataFrame(data=_np.linspace(tstart, tcoal, Npts), columns=['time'])  # Initial column, 100 values, -10 - 0
    
    GMcc3 = _ac.g*Mc/_ac.c**3
    df['fgw']  = 1/_ac.pi * _np.power(5/256 * 1/(tcoal-df.time), 3/8) * GMcc3**(-5/8)
    df = df[df.fgw < f_max]  # Cut the REALLY wrong bits...
    
    df['dfdt'] = 96/5 * _ac.pi**(8/3) * GMcc3**(5/3) * _np.power(df.fgw, 11/3)
    df['worb'] = df.fgw/2 * _ac.pi2  # worb = forb * 2pi = f_gw/2 * 2pi
    df['aorb'] = _np.power((_ac.g*mt) / _np.square(df.worb), 1/3)  # aorb^3 = G Mt / w^2
    
    df['vorb1'] = df.worb * df.aorb * m2/mt / _ac.c
    df['vorb2'] = df.worb * df.aorb * m1/mt / _ac.c
    
    df['ampl'] = 4/dist * _ac.g/_ac.c**4 * mu * _np.square(df.aorb) * _np.square(df.worb)
    h_phase    = 2*df.worb*(df.time-tcoal)
    df['hpl']  = df.ampl * (1+cosi**2)/2 * _np.cos(h_phase)
    df['hcr']  = df.ampl * cosi          * _np.sin(h_phase)
    df['h']    = Fplcr * (df.hpl + df.hcr)
    
    if verbosity>0:
        print('f_low, f_high (Hz):          ', df.fgw.iloc[0], df.fgw.iloc[-1])
        print('a_min (km):                  ', a_min/1000)
        print('f_isco, f_max (Hz):          ', f_isco, f_max)
        
        if verbosity>8: print(df)
    
    
    return df


def cbc_waveform_frequency(m1,m2, dist,cosi, f_low,f_high, Npts, risco_fac=1.5, Fplcr=2/5, verbosity=0):
    """Compute a simple, Newtonian(!) compact-binary-coalescence waveform in the frequency domain for a given
    range of frequencies.
    
    Parameters:
      m1 (float):         Mass of binary component 1 (kg).
      m2 (float):         Mass of binary component 2 (kg).
      dist (float):       Distance of binary (m).
      cosi (float):       Cosine of the inclination angle: +/-1 = face on, 0 = edge-on.
      f_low (float):      Lower cut-off frequency (Hz).
      f_high (float):     Upper cut-off frequency (Hz).
      Npts (float):       Number of data points (-).
      risco_fac (float):  R_isco = R_star * risco_fac (-, defaults to 1.5); used to compute f_max.
      Fplcr (float):      The mean value of F_+ and F_x: <F> ~ sqrt(<F_+^2 + F_x^2>),
                          which includes the inclination.  Defaults to 2/5.
      verbosity (int):    Verbosity level, defaults to 0: no output.
    
    Returns:
      (pd.df):  Pandas dataframe containing variables, including:
                - fgw:            GW frequency in Hz;
                - htilde:         h~ in 1/Hz;
                - htilde_pSqrtHz: h~ in 1/sqrt(Hz)  (= 2 * htilde * sqrt(fgw)).
    """
    
    # Create DataFrame with the frequency range and call cbc_waveform_frequency_array() to fill it:
    df = _pd.DataFrame(data=_np.logspace(_np.log10(f_low), _np.log10(f_high), Npts), columns=['fgw'])  # Initial column, Npts rows
    
    df = cbc_waveform_frequency_array(df, m1,m2, dist,cosi, risco_fac=1.5, Fplcr=2/5, verbosity=0)
    
    return df


def cbc_waveform_frequency_array(df, m1,m2, dist,cosi, risco_fac=1.5, Fplcr=2/5, verbosity=0):
    """Compute a simple, Newtonian(!) compact-binary-coalescence waveform in the frequency domain for a given
    range of frequencies.
    
    Parameters:
      df (pd.df):         Pandas DataFrame containing frequency array (Hz) in column 'fgw'.
      m1 (float):         Mass of binary component 1 (kg).
      m2 (float):         Mass of binary component 2 (kg).
      dist (float):       Distance of binary (m).
      cosi (float):       Cosine of the inclination angle: +/-1 = face on, 0 = edge-on.
      risco_fac (float):  R_isco = R_star * risco_fac (-, defaults to 1.5); used to compute f_max.
      Fplcr (float):      The mean value of F_+ and F_x: <F> ~ sqrt(<F_+^2 + F_x^2>),
                          which includes the inclination.  Defaults to 2/5.
      verbosity (int):    Verbosity level, defaults to 0: no output.
    
    Returns:
      (pd.df):  Pandas dataframe containing variables, including:
                - fgw:            GW frequency in Hz (input);
                - htilde:         h~ in 1/Hz;
                - htilde_pSqrtHz: h~ in 1/sqrt(Hz)  (= 2 * htilde * sqrt(fgw)).
    """
    
    mt = m1+m2
    Mc = (m1*m2/mt**(1/3))**(3/5)
    
    R1 = radius_BH_NS_from_mass(m1)  # Estimate/compute the BH/NS radius from the object mass
    R2 = radius_BH_NS_from_mass(m2)
    a_min = R1+R2
    
    f_isco  = _ac.c**3 / (6**(3/2) * _ac.pi * _ac.g * mt)
    # f_max = _ac.c**3 / (3**(3/2) * _ac.pi * _ac.g * mt)       # Roughly matches PRX 6 041015 (2016)
    f_max = 1/_ac.pi * _np.sqrt(_ac.G * mt / (risco_fac*a_min)**3)  # f_max based on 1.5 x (sum of radii) and Kepler - same as above, but allows NSs
    
    
    # Frequency domain (see Maggiore Eqs. 4.43-37, p.174):
    h_sq = 5/(24 * _ac.pi**(4/3) * dist**2 * _ac.c**3) * (_ac.G * Mc)**(5/3) / _np.power(df.fgw, 7/3)  # |h(f)|^2 (Maggiore Eq.4.370, p.231)
    # h_sq *= 1.665  # Hack that would give nicer matches! - i.e., 2/5 -> 2/3 or sqrt(2/5)! in the line below
    df['htilde']         = Fplcr * _np.sqrt(h_sq)                # [h] = 1/Hz - Maggiore, Table 7.1, Eq.7.181
    df['htilde_pSqrtHz'] = 2     * df.htilde * _np.sqrt(df.fgw)  # [h] = 1/sqrt(Hz) - for (plot) comparison to ASD; see PRX 6, 041015 (2016), Fig.1
    
    # df['Psi_pl'] = _ac.pi2 * df.fgw * 1  -  0  - _ac.pio4  \
    #     +  3/4 * _np.power(_ac.G * Mc/_ac.c**3 * 8*_ac.pi * df.fgw, -5/3)
    # df['Psi_cr'] = df.Psi_pl + _ac.pio2
    # df['htilde_pl'] = df.htilde * (1+cosi**2)/2 * _np.exp( df.Psi_pl * 1j )
    # df['htilde_cr'] = df.htilde * cosi          * _np.exp( df.Psi_cr * 1j )
    # df['htilde_tot'] = _np.abs(_np.real(df.htilde_pl + df.htilde_cr)/2)  # _np.cos(df.Psi_pl)
    
    
    df = df[df.fgw < f_max]  # Cut the REALLY wrong bits...
    f_low  = df.fgw.iloc[0]
    f_high = df.fgw.iloc[-1]
    
    if verbosity>0:
        print('f_low, f_high (Hz):          ', f_low, f_high)
        print('a_min (km):                  ', a_min/1000)
        print('f_isco, f_max (Hz):          ', f_isco, f_max)
        print('h_start, h_end (1/Hz):       ', df.htilde.iloc[0], df.htilde.iloc[-1])
        print('h_start, h_end (1/sqrt Hz):  ', df.htilde_pSqrtHz.iloc[0], df.htilde_pSqrtHz.iloc[-1])
        
        if verbosity>8: print(df)
    
    return df


def noise_curve(Npts, f_low,f_high, Len, P_L,lam_L, eta_pd, M_mir, R_in,R_end, PRfac, no_FP=False,
                adhoc_lowf=False, verbosity=0):
    
    """Compute a simplified noise curve for a gravitational-wave interferometer.
    
    Parameters:
      Npts (float):     Number of data points (-).
      f_low (float):    Lower frequency cut off (Hz).
      f_high (float):   Higher frequency cut off (Hz).
    
      Len (float):      Length of interferometer arms (m).
      P_L (float):      Laser power (W).
      lam_L (float):    Laser wavelength (m).
      eta_pd (float):   Efficiency of photodetector (fraction 0-1).
    
      M_mir (float):    Mass of the end mirror/test mass (kg).
      R_in (float):     Reflectivity of the input mirrors of the (main) Fabry-Perot cavity (fraction 0-1).
      R_end (float):    Reflectivity of the end mirrors of the interferometer (fraction 0-1).
    
      PRfac (float):    Factor for power recycling (factor 1-...).
      no_FP (bool):     Do NOT use a Fabry-Perot cavity (optional; defaults to False, i.e. DO use a FPC).

      adhoc_lowf (bool):  Add an ad-hoc low-f term to mimic termal/Newtonian/seismic noise (defaults to False).
      verbosity (int):    Verbosity level, defaults to 0: no output.
    
    Returns:
      (pd.df):  Pandas.DataFrame containing noise strains, with columns:
                - asd: the total amplitude spectral density, in strain/sqrt(Hz);
                - asd_shot:  the shot-noise component of the ASD, in strain/sqrt(Hz);
                - asd_rad:   the radiation-pressure component of the ASD, in strain/sqrt(Hz);
                - asd_lowf:  the ad-hoc low-frequency component of the ASD, in strain/sqrt(Hz), if desired.
    """
    
    # Create DataFrame:
    df = _pd.DataFrame(data=_np.logspace(_np.log10(f_low), _np.log10(f_high), Npts), columns=['fgw'])  # Initial column, Npts rows
    
    if no_FP:
        Fin        = _ac.pi/2
        Leff_L     = 1
        f_pole     = 0
        f_pole_fac = 1
    else:
        Fin        = _ac.pi * _np.sqrt(R_in*R_end) / (1 - R_in*R_end)   # Finesse
        Leff_L     = 2*Fin/_ac.pi
        f_pole     = _ac.c / (4 * Fin * Len)                         # Pole frequency
        f_pole_fac = _np.sqrt( 1 + _np.square(df.fgw/f_pole) )
        
        
    df['asd'] = 0  # Most important column first - assign proper values below
    
    df['asd_shot'] = \
        1/(8*Fin*Len) * _np.sqrt( (4*_ac.pi*_ac.h_bar*_ac.c * lam_L) / (eta_pd * PRfac * P_L) ) * f_pole_fac
    
    df['asd_rad'] = 16*_np.sqrt(2) * Fin / (M_mir * Len * _np.square(_ac.pi2*df.fgw)) \
        * _np.sqrt((_ac.h_bar * P_L * PRfac)/(_ac.pi2 * lam_L * _ac.c)) / f_pole_fac
    
    df['asd'] = df.asd_shot + df.asd_rad
    
    if adhoc_lowf:
        # Ad-hoc low-f term, (strongly!) adapted from seismic noise Bader PhD thesis, Eq. 1.2.15:
        seismic_isolation = 1e-4  # 10^10 needed for h ~ 1e-23 m/sqrt(Hz) @10Hz
        alpha = 1e-6  # 1e-6 - 1e-9 m Hz^(3/2)
        pow = 6       # {1e-4, 1e-6, 6} looks roughly like PRX 6 041015 (2016), Fig.1 curves for O1 (and O2?)
        df['asd_lowf'] = alpha / _np.power(df.fgw, pow) / Len * seismic_isolation
        df.asd += df.asd_lowf
    
    if verbosity > 2:
        print()
        print('Len:            ', Len/1000, 'km')
        print('P_L:            ', P_L, 'W')
        print('lam_L:          ', lam_L*1e9, 'nm')
        print('eta_pd:         ', eta_pd)
        print()
        print('M_mir:          ', M_mir, 'kg')
        print('R_in:           ', R_in)
        print('R_end:          ', R_end)
        print()
        print('PRfac:          ', PRfac)
        print()
        
    if verbosity > 1:
        print()
        print('Finesse:        ', Fin)
        print('Leff/L:         ', Leff_L)
        print('f_pole:         ', f_pole)
        print()
        print('min ASD_shot:   ', df.asd_shot.min(), '/sqrt(Hz)')
        print('min ASD_rad:    ', df.asd_rad.min(),  '/sqrt(Hz)')
        print('min ASD:        ', df.asd.min(),      '/sqrt(Hz)')
        print('min ASD:        ', df.asd.min()*10,   '@ 100Hz')
        print()
        # print('MI ASD: ', 1/(_ac.pi*Len) * )
    
    return df

    
def isco_radius_from_mass(mass, spin=0, as_km=False):
    """Compute the black-hole ISCO GW radius from its mass and spin.
    
    Parameters:
      mass (float):  mass of the black hole (Mo).
      spin (float):  spin of the black hole (-1 - +1, defaults to 0).
      as_km (bool):  the result is expressed in km, else in terms of the BH mass M (defaults to False).
    
    Returns:
      (float):  ISCO radius in either km or M, as desired.
    """
    
    scalar_input = True
    mass = _np.asarray(_np.copy(mass+spin*0), dtype=float)  # Copy and typecast to numpy.ndarray - ensure float!
    spin = _np.asarray(_np.copy(mass*0+spin), dtype=float)  # Copy and typecast to numpy.ndarray - ensure float!
    if mass.ndim == 0:
        mass = mass[None]  # Makes mass a 1D array.  Comment: use np.newaxis instead?
    else:
        scalar_input = False
    if spin.ndim == 0:
        spin = spin[None]  # Makes spin a 1D array.  Comment: use np.newaxis instead?
    else:
        scalar_input = False
    
    z1 = 1 + _np.power(1 - _np.square(spin),_ac.c3rd) * \
        ( _np.power(1+_np.abs(spin),_ac.c3rd) + _np.power(1-_np.abs(spin),_ac.c3rd) )
    z2 = _np.sqrt(3*_np.square(spin) + _np.square(z1))
    
    r_isco = _np.zeros_like(mass+spin)
    sel = spin<0
    r_isco[sel]  = mass[sel] * (3+z2[sel] - _np.sqrt((3-z1[sel])*(3+z1[sel] + 2*z2[sel])) )  # Expressed in M
    sel = spin>=0
    r_isco[sel]  = mass[sel] * (3+z2[sel] + _np.sqrt((3-z1[sel])*(3+z1[sel] + 2*z2[sel])) )  # Expressed in M
    
    if as_km: r_isco *= _ac.G*_ac.Mo/_ac.c**2 / 1000  # Convert from M to m to km
    
    if scalar_input:
        return float(_np.squeeze(r_isco))  # Array -> scalar (float)
    
    return r_isco


def isco_frequency_from_isco_radius(mass, r_isco, as_km=False):
    """Compute the ISCO GW frequency from the mass and ISCO radius expressed in M.
    
    Parameters:
      mass (float):    mass of the object (Mo).
      r_isco (float):  ISCO radius of the object (in units of M, the BH mass).
      as_km (bool):    the ISCO radius is expressed in km, else in terms of the BH mass M (defaults to False).
    
    Returns:
      (float):  ISCO GW frequency in Hertz.
    """
    
    if not as_km: r_isco *= _ac.G*_ac.Mo/_ac.c**2 / 1000  # Convert from M to m to km
    
    f_isco_gw = 2 * _ac.c**3 / (2*_ac.pi * _np.power(r_isco/mass,1.5) * _ac.g*mass*_ac.m_sun)
    
    return f_isco_gw


def isco_frequency_from_mass(mass, spin=0):
    """Compute the ISCO GW frequency from the mass and ISCO radius expressed in M.
    
    Parameters:
      mass (float):    mass of the object (Mo).
      spin (float):    spin of the black hole (-1 - +1, defaults to 0).
    
    Returns:
      (float):  ISCO GW frequency in Hertz.
    """
    
    r_isco = isco_radius_from_mass(mass, spin, as_km=False)  # R_isco in terms of M
    f_isco_gw = 2 * _ac.c**3 / (2*_ac.pi * _np.power(r_isco/mass,1.5) * _ac.g*mass*_ac.m_sun)
    
    return f_isco_gw


def radius_BH_NS_from_mass(mass):
    """Estimate the radius of a (non-spinning) black hole or neutron star from its mass.
    
    Parameters:
      mass (float):  mass of the object (kg).
    
    Returns:
      (float):  radius of the object (m).
    
    
    Note: this function simply returns:
      - 11.5 km as the radius of a NS if mass < 3.9Mo;
      - the Schwarzschild radius if mass >= 3.9Mo.
    
    Rationale:
      - all NSs have radii of ~11.5km here: https://arxiv.org/abs/1205.6871;
      - Rs ~ 11.5km for a 3.9Mo BH.
    """
    
    if mass < 3.9*_ac.sun_m:         # R_s ~ 11.5km for BH of 3.9Mo
        rad = 11.5*_ac.km            # R_NS (m)
    else:
        rad = 2*_ac.G*mass/_ac.c**2  # R_BH (m)
    return rad



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: EUPL-1.2
#  
#  Copyright (c) 2022-2024  Marc van der Sluys - Nikhef/Utrecht University - marc.vandersluys.nl
#   
#  This file is part of the sluyspy Python package:
#  Marc van der Sluys' personal Python modules.
#  See: https://github.com/MarcvdSluys/sluyspy
#   
#  This is free software: you can redistribute it and/or modify it under the terms of the European Union
#  Public Licence 1.2 (EUPL 1.2).  This software is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the EU Public Licence for more details.  You should have received a copy of the European
#  Union Public Licence along with this code.  If not, see <https://www.eupl.eu/1.2/en/>.


"""Solar-panel functions for the sluyspy package"""


import pandas as _pd
import pytz as _tz
import solarenergy as _se
from sluyspy import env as _env

# My computing environment:
_env = _env.environment()

# My solar-panel specs:
specs = _se.read_solar_panel_specs()


def read_detailed_log(file_name='Current/detailed-log.csv', last_only=None, header=None, sun_position=False, rem_cols=True, no_p0rows=False, no_elec=False, no_cond=True):
    """Read solar-panel detailed-log.csv and select the useful data.
    
    Parameters:
      file_name (str):      Name of the input file (optional; default=None -> detailed-log.csv).
      header (int):         Number of lines to consider as header and skip (optional; default: 0).
      last_only (int):      Number of last lines to read.  This tails the last lines into a separate file and reads that.
      sun_position (bool):  Add (compute) Sun position to the data (slow! - optional; default: False).
      rem_cols (bool):      Remove default columns (constants, duplicates, uninteresting).
      no_p0rows (bool):     Remove rows without power (P=0; Relay=closed; Cond!=OK).
      no_elec (bool):       Remove electricity details (some P,V,I, f).
      no_cond (bool):       Remove condition/status columns (Cond, Relay, Tinv).
      
    Returns:
      (pandas.DataFrame):  DataFrame containing solar-panel data.
    """
    
    col_names = ['date', 'time', 'SN1', 'Type', 'SN2', 'Pdc', 'x0', 'Idc', 'x1', 'Vdc', 'x2', 'Pac', 'x3', 'x4', 'Iac', 'x5', 'x6', 'Vac',
                 'x7', 'x8', 'Pdc2', 'Pac2', 'x9', 'Eday', 'Etot', 'Freq', 't_oper', 't_feedin', 'Btooth', 'Cond', 'Relay', 'Tinv']
    
    # Issue: detailed-log.csv becomes very long -> reading is slow.  Tail the last 3000 lines/minutes (2+
    # days) to a shorter file and read that (currently saves a factor 17 on CPU time (2022-11-12)):
    if last_only is not None:
        from pathlib import Path
        import sluyspy.system as ssys
        
        tmpfile = ssys.temp_file_name(_env.sp_dir+'Current',  '.detailed-log-last', 'csv')
        
        ssys.tail_file(_env.sp_dir+file_name, tmpfile, last_only)  # Copy last last_only lines to tmpfile
        df = _pd.read_csv(tmpfile, header=header, sep=r'\s*,\s*', engine='python', names=col_names)
        
        Path.unlink(Path(tmpfile))  # Remove temporary file
        
    else:
        df = _pd.read_csv(_env.sp_dir+file_name, header=header, sep=r'\s*,\s*', engine='python', names=col_names)
    
    # Remove unwanted columns:
    if rem_cols:
        del df['SN1'], df['SN2'], df['Type']  # Not interesting - constant
        del df['Pdc2'], df['Pac2']            # Not interesting - duplicates
        del df['x0'], df['x1'], df['x2'], df['x3'], df['x4'], df['x5'], df['x6'], df['x7'], df['x8'], df['x9']  # x_i are always 0
        
        del df['Eday'], df['t_oper'], df['t_feedin'], df['Btooth']  # Not interested in now  # , df['Etot']
        
    
    # Remove rows witout power (P=0; Relay=closed; Cond!=OK):
    if no_p0rows:
        df = df[df['Relay'] == 'Closed']  # Only keep data when the relay is closed (i.e., providing a current/energy)
        df = df[df['Cond']  == 'Ok']      # Only keep data when the condition is 'OK'
        df = df[df['Pdc']  > 0]           # Only keep data if there is power
        df = df[df['Pac']  > 0]           # Only keep data if there is power
        
        # df = df[df['Pac']  < 2990]        # Only keep data for P < 3kW, since 3kW may indicate >3kW
        # df = df[df['Tinv'] > 0]           # Only keep data if we have a temperature (of the inverter!)
        
        # No longer needed:
        del df['Relay'], df['Cond'], df['Tinv']
        
    elif no_cond:
        del df['Cond'], df['Relay'], df['Tinv']
    
    # Slightly more accurate: ~3-4 significant digits -> 4-5:
    df['Pdc'] = df['Idc'] * df['Vdc']
    df['Pac'] = df['Iac'] * df['Vac']
    
    if no_elec:
        del df['Pdc'], df['Idc'], df['Vdc'], df['Iac'], df['Vac'], df['Freq']

    
    
    # Combine date and time columns to a timezone-aware datetime:
    df['date'] = _pd.DatetimeIndex(_pd.to_datetime(df['date']+' '+df['time']))
    df = df.rename(columns={'date': 'dtm'})  # Rename column date -> dtm
    # df['dtm'] = df['dtm'].dt.round('min')
    mytz = _tz.timezone(specs.tz)
    df.dtm = df.dtm.dt.tz_localize(mytz)  # Timezone naive -> timezone aware.  Indicate that the existing times are in mytz, without conversion.
    
    del df['time']  # Remove column 'time'
    
    # print(df['dtm'])
    
    # Compute Sun position (uses SolTrack behind the scenes):
    if sun_position:
        df['sunAz'],df['sunAlt'],df['sunDist'] = \
            _se.sun_position_from_datetime(specs.geo_lon,specs.geo_lat, df['dtm'])
        
    return df


def write_day_file(df, subdir='', prefix=specs.name, ext='csv', verbosity=0):
    """Write one day of (e.g.) minutely solar-panel data to daily file named 'subdir/prefix-yyyymmdd.ext'.
    
    Parameters:
      df (pandas.DataFrame):  Pandas DataFrame with columns datetime, Pac and Etot.
      subdir (str):           Subdirectory to save file in, with trailing slash.
      prefix (str):           File name prefix (defaults to name of the plant, followed by a hyphen).
      ext (str):              File extension to use.
      verbosity (int):        Verbosity (0-2; defaults to 0).
    """
    
    if len(df) < 1:
        if verbosity > 0:  print('No data available, no file created.')
        return
    
    
    # Want separate date and time columns:
    df['date'] = df.dtm.dt.date
    df['time'] = df.dtm.dt.time
    
    date = df.loc[:, 'dtm'].dt.date.iloc[0].strftime('%Y%m%d')  # .loc[0, ''] doesn't work, since row 0 does not exist
    out_file_name = _env.sp_dir+subdir+prefix+'-'+date+'.'+ext
    if verbosity > 0:  print('Saving data to '+out_file_name)
        
    # Open file and write header:
    out_file = open(out_file_name, 'w')
    out_file.write('sep=;\n')
    out_file.write('Version CSV1|Tool SBFspot3.3.1 (Linux)|Linebreaks CR/LF|Delimiter semicolon|Decimalpoint dot|Precision 3\n')
    out_file.write('\n')
    out_file.write(';SN: '+specs.inv_sn+';SN: '+specs.inv_sn+'\n')
    out_file.write(';'+specs.inv_model+';'+specs.inv_model+'\n')
    out_file.write(';'+specs.inv_sn+';'+specs.inv_sn+'\n')
    out_file.write(';Total yield;Power\n')
    out_file.write(';Counter;Analog\n')
    out_file.write('yyyy-MM-dd;HH:mm:ss;kWh;kW\n')
    
    for row in df.iterrows():
        ser = row[1]
        if verbosity > 1:  print('%s;%s;%0.3f;%5.3f' % (ser.date, ser.time, ser.Etot, ser.Pac))
        
        out_file.write('%s;%s;%0.3f;%5.3f\n' % (ser.date, ser.time, ser.Etot, ser.Pac/1e3))  # Pac: W -> kW
        
    out_file.close()
    
    return    





#!/bin/env python
# -*- coding: utf-8 -*-

""" test_gws.py:  
    2024-04-18, MvdS: initial version.
"""


import colored_traceback
colored_traceback.add_hook()

import numpy as np
import pandas as pd


def main():
    """Main function."""

    test_isco()
    exit(0)

    
def test_isco():
    import sluyspy.gws as gw
    
    df = pd.DataFrame(data=np.linspace(0,31,32), columns=['mass'])
    df.loc[0,'mass'] = 0.1
    df['f_isco'] = gw.isco_frequency_from_mass(df.mass, 0)
    print(df)
    
    df = pd.DataFrame(data=np.linspace(-1,1,11), columns=['spin'])
    df['f_isco'] = gw.isco_frequency_from_mass(10, df.spin)
    print(df)
    
    df['mass']   = np.linspace(5,55,11)
    df['f_isco'] = gw.isco_frequency_from_mass(df.mass, df.spin)
    print(df)
    
    return


if __name__ == '__main__':
    main()

